/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


#include <xacml_client.h>
#include <xacml_io_ssl.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <errno.h>
#include <grp.h>


/* B: prototypes */

#include "pep_obligation_handlers.h"
#include "xacml_attr_obl_profile.h"

#include <lcmaps/lcmaps_log.h>
#include <lcmaps/lcmaps_cred_data.h>



static int oh_uidgid_fired        = 0;
static int oh_secondarygids_fired = 0;
static int oh_username_fired      = 0;


/*
int pep_construct_request_subject (xacml_request_t   request,
                           char *            subject_dn,
                           char *            issuer_dn,
                           int               subject_certificate_serial_number,
                           int               subject_ca_serial_number,
                           char *            subject_vo[],
                           int               subject_vo_cnt,
                           char *            voms_signing_subject[],
                           int               voms_signing_subject_cnt,
                           char *            voms_signing_issuer[],
                           int               voms_signing_issuer_cnt,
                           char *            voms_dns_port[],
                           int               voms_dns_port_cnt,
                           char *            subject_voms_fqan[],
                           int               subject_voms_fqan_cnt,
                           char *            subject_voms_primary_fqan,
                           char *            ca_policy_oid[],
                           int               ca_policy_oid_cnt,
                           char *            cert_chain
                          );
int pep_construct_request_resource (xacml_request_t   request,
                                    char *            resource_node_type,
                                    char *            host_dns_name
                                   );
int pep_construct_request_action   (xacml_request_t   request,
                                    char *            action_type
                                   );
*/
/* E: prototypes */


/*************************** B: Obligation Handlers *********************************/
int
default_handler(
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    char * logstr = "default_handler";
    int i;

    lcmaps_log (1, "%s: Error: Unknown/unregistered obligation received, must fail authorized! Obligation ID: %s\n", logstr, obligation_id);
    for (i = 0; attribute_ids[i] != NULL; i++)
    {
        lcmaps_log (5, "%s: %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);
    }

    return 1;
}




int
UIDGID_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    char * logstr = "UIDGID_handler";
    int rc = 0;
    int i  = 0;
    uid_t uid = -1;
    gid_t gid = -1;

    lcmaps_log_debug (3, "%s: Got obligation %s\n", logstr, obligation_id);
    setOHUIDGIDFired(1);

    for (i = 0; attribute_ids[i] != NULL && rc == 0; i++)
    {
        lcmaps_log_debug (3, "%s: %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);

        /* posix-uid attribute */
        if (strcasecmp (attribute_ids[i], XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID) == 0)
        {
            uid = atoi (values[i]);
            lcmaps_log_debug (2, "%s: Adding UID :  %d\n", logstr, uid);
            rc = addCredentialData(UID, &uid);
        } 

        /* posix-gid attribute */
        else if (strcasecmp (attribute_ids[i], XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID) == 0)
        {
            gid = atoi (values[i]);
            lcmaps_log_debug (2, "%s: Adding GID :  %d\n", logstr, gid);
            rc = addCredentialData(PRI_GID, &gid);
        }
        else
        {
            lcmaps_log (1, "%s: Error: Unexpected %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);
            return 1;
        }
    }

    return rc;
}


int
SecondaryGIDs_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    char * logstr = "SecondaryGIDs_handler";
    int rc = 0;
    int i  = 0;
    gid_t gid = -1;

    lcmaps_log_debug (3, "%s: Got obligation %s\n", logstr, obligation_id);
    setOHSecondaryGIDsFired(1);

    for (i = 0; attribute_ids[i] != NULL && rc == 0; i++)
    {
        lcmaps_log_debug (3, "%s: %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);

        /* posix-gid attribute */
        if (strcasecmp (attribute_ids[i], XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID) == 0)
        {
            gid = atoi (values[i]);
            lcmaps_log_debug (2, "%s: Adding GID :  %d\n", logstr, gid);
            rc = addCredentialData(SEC_GID, &gid);
        }
        else
        {
            lcmaps_log (1, "%s: Error: Unexpected %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);
            return 1;
        }
    }

    return rc;
}


int
Username_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    char * logstr = "Username_handler";
    int rc = 0;
    int i  = 0;
    int j  = 0;
    uid_t uid = -1;
    gid_t gid = -1;
    int cnt_sec_gid = 0;
    gid_t * sec_gid = NULL;


    lcmaps_log (5, "%s: Got obligation %s\n", logstr, obligation_id);
    setOHUsernameFired(1);

    for (i = 0; attribute_ids[i] != NULL && rc == 0; i++)
    {
        lcmaps_log (5, "%s: %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);

        /* Username attribute */
        if (strcasecmp (attribute_ids[i], XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME) == 0)
        {
            const char * username = values[i];
            struct passwd * pw = NULL; 

            lcmaps_log (5, "%s: Found username :  %s\n", logstr, username);

            if (! (username && (strlen(username) > 0)))
            {
                lcmaps_log (1, "%s: Error: Couldn't find the Username value with the attribute identifier. This is a protocol error.\n");
                return 1;
            }

            pw = getpwnam(username);
            if (!pw)
            {
                lcmaps_log (1, "%s: Error: Couldn't find the username '%s' in the password file.\n", logstr, username);
                return 1;
            } 
            else
            {
                uid = pw -> pw_uid;
                gid = pw -> pw_gid;

                lcmaps_log (5, "%s: Adding UID :  %d\n", logstr, uid);
                lcmaps_log (5, "%s: Adding GID :  %d\n", logstr, gid);
                rc = addCredentialData(UID, &uid);
                rc = addCredentialData(PRI_GID, &gid);

                /*
                 * Retrieve secondary group id's
                 */
                if (get_gidlist(username, &cnt_sec_gid, &sec_gid)==0)
                {
                    for (j = 0; j < cnt_sec_gid; j++)
                    {
                        lcmaps_log (5, "%s: Adding SGID :  %d\n", logstr, sec_gid[j]);
                        addCredentialData(SEC_GID, &(sec_gid[j]));
                    }
                    free(sec_gid);
                }
            }
        }
        else
        {
            lcmaps_log (1, "%s: Error: Unexpected %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);
            return 1;
        }
    }

    return rc;
}




int         
AFSToken_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{   
    int i;
    char * logstr = "AFSToken_handler";
    lcmaps_log (5, "AFSToken: Got obligation %s\n", logstr, obligation_id);
    
    for (i = 0; attribute_ids[i] != NULL; i++)
    {
        lcmaps_log (5, "%s: %s [%s] = %s\n", logstr, attribute_ids[i], datatypes[i], values[i]);
    }
    return 1;
}
/*************************** E: Obligation Handlers *********************************/



/************************************************************
Function:     pep_register_obligation_handlers
Parameters:
              xacml_request_t   request
Description:  
              This function will register the obligations that the PEP understands
              and it will register it with a handler function for that obligation.
Parameters:
              xacml_request_t   request
return values:
      must be 0
************************************************************/
int pep_register_obligation_handlers (xacml_request_t request)
{

    /************ B: Register Obligation Handlers *********/

    /* Obligation: UID & GID (should feature two attributes) */
    xacml_request_add_obligation_handler(
            request,
            UIDGID_handler,
            NULL,
            XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID);

    /* Obligation: Multiple Secondary GIDs in multiple attributes (requires: UIDGID or Username) */
    xacml_request_add_obligation_handler(
            request,
            SecondaryGIDs_handler,
            NULL,
            XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs);

    /* Obligation: AFS Token (requires: UIDGID or Username) */
    xacml_request_add_obligation_handler(
            request,
            AFSToken_handler,
            NULL,
            XACML_ATTR_PROFILE_NS_URL_OBLIGATION_AFSToken);

    /* Obligation: Username (should match the UIGGID.uid attribute value) */
    xacml_request_add_obligation_handler(
            request,
            Username_handler,
            NULL,
            XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username);

#ifdef STAKEHOLDER_OSG

    /* Obligation: Path restriction (requires: UIDGID or Username) */
    xacml_request_add_obligation_handler(
            request,
            RootAndHomePaths_handler,
            NULL,
            XACML_ATTR_PROFILE_NS_URL_OBLIGATION_RootAndHomePaths);

    /* Obligation: Storage priorities (requires: UIDGID or Username) */
    xacml_request_add_obligation_handler(
            request,
            StorageAccessPriority_handler,
            NULL,
            XACML_ATTR_PROFILE_NS_URL_OBLIGATION_StorageAccessPriority);

    /* Obligation: Storage Access Permissions (requires: (UIDGID or Username) and RootAndHomePaths ) */
    xacml_request_add_obligation_handler(
            request,
            AccessPermissions_handler,
            NULL,
            XACML_ATTR_PROFILE_NS_URL_OBLIGATION_AccessPermissions);
#endif /* STAKEHOLDER == "OSG" */

/* 
    xacml_request_add_obligation_handler(
            request,
            local_user_name_handler,
            NULL,
            "urn:globus:local-user-name:obj");
*/
    /*** Default handler: this handler will be used for fault handling of the obligation that are not supported ***/
    xacml_request_add_obligation_handler(
            request,
            default_handler,
            NULL,
            NULL);
    /************ E: Register Obligation Handlers *********/


    return 0;
}



/************************************************************
Function:     pep_construct_request_subject
Parameters:
Description:  
Parameters:
return values:
      must be 0
************************************************************/
int pep_construct_request_subject (xacml_request_t   request,
                           char *            subject_dn,
                           char *            issuer_dn,
                           int               subject_certificate_serial_number,
                           int               subject_ca_serial_number,
                           char *            subject_vo[],
                           int               subject_vo_cnt,
                           char *            voms_signing_subject[],
                           int               voms_signing_subject_cnt,
                           char *            voms_signing_issuer[],
                           int               voms_signing_issuer_cnt,
                           char *            voms_dns_port[],
                           int               voms_dns_port_cnt,
                           char *            subject_voms_fqan[],
                           int               subject_voms_fqan_cnt,
                           char *            subject_voms_primary_fqan,
                           char *            ca_policy_oid[],
                           int               ca_policy_oid_cnt,
                           char *            cert_chain
                          )
{
    char *       logstr = "pep_construct_request_subject()";
    int          i = 0;
    char         buf[256];
/*
    char *       buf = NULL;
    buf = malloc (sizeof (char) * 256);
*/


    /************ B: Subject *********/

    /* Required operation */

    if (! subject_dn)
    {
        lcmaps_log (0, "%s: Error: No subject DN found, this element is mandatory\n", logstr);
        return 1;
    }

    xacml_request_set_subject(
            request,
            subject_dn);

    xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ID,
            XACML_DATATYPE_STRING,
            "",
            subject_dn);

    if (issuer_dn)
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ISSUER,
            XACML_DATATYPE_STRING,
            "",
            issuer_dn);

    if (subject_certificate_serial_number != 0)
    {
        snprintf (buf, 255, "%d", subject_certificate_serial_number);
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_CERTIFICATE_SERIAL_NUMBER,
            XACML_DATATYPE_STRING,
            "",
            buf);
    }

    if (subject_ca_serial_number != 0)
    {
        snprintf (buf, 255, "%d", subject_ca_serial_number);
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_CA_SERIAL_NUMBER,
            XACML_DATATYPE_INTEGER,
            "",
            buf);
    }


    for (i = 0; i < subject_vo_cnt; i++)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_VO,
            XACML_DATATYPE_STRING,
            "",
            subject_vo[i]);

    }

    for (i = 0; i < voms_signing_subject_cnt; i++)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_SUBJECT,
            XACML_DATATYPE_STRING,
            "",
            voms_signing_subject[i]);
    }

    for (i = 0; i < voms_signing_issuer_cnt; i++)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_ISSUER,
            XACML_DATATYPE_STRING,
            "",
            voms_signing_issuer[i]);
    }

    for (i = 0; i < voms_dns_port_cnt; i++)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_VOMS_DNS_PORT,
            XACML_DATATYPE_STRING,
            "",
            voms_dns_port[i]);
    }

    for (i = 0; i < subject_voms_fqan_cnt; i++)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_FQAN,
            XACML_DATATYPE_STRING,
            "",
            subject_voms_fqan[i]);
    }

    if (subject_voms_primary_fqan)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_PRIMARY_FQAN,
            XACML_DATATYPE_STRING,
            "",
            subject_voms_primary_fqan);
    }

    for (i = 0; i < ca_policy_oid_cnt; i++)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_CA_POLICY_OID,
            XACML_DATATYPE_STRING,
            "",
            ca_policy_oid[i]);
    }

    if (cert_chain)
    {
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN,
            XACML_DATATYPE_STRING,
            "",
            cert_chain);
    }
    /************ E: Subject *********/


    return 0;
}


int pep_construct_request_resource (xacml_request_t   request,
                                    char *            resource_type,
                                    char *            host_dns_name,
                                    char *            resource_id,
                                    char *            resource_x509_id,
                                    char *            resource_x509_issuer
                                   )
{
    xacml_resource_attribute_t ra;

    /************ B: Resource *********/

    /* Node type (WN, CE, SE, RB) */
    /* Note: The Resource Type is names by the the 'Resource-id'.
     *       this was the XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE.
     *       This is to use more of the already standardized attribute namespaces.
     */
    if (resource_type)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID,
                XACML_DATATYPE_STRING,
                "",
                resource_type);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }

    /* DNS Host name */
    if (host_dns_name)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME,
                XACML_DATATYPE_STRING,
                "",
                host_dns_name);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }

    /* resource-id */
    if (resource_id)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID,
                XACML_DATATYPE_STRING,
                "",
                resource_id);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }

    /* resource-x509-id */
    if (resource_x509_id)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ID,
                XACML_DATATYPE_STRING,
                "",
                resource_x509_id);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }
    /************ E: Resource *********/

    /* resource-x509-issuer */
    if (resource_x509_issuer)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ISSUER,
                XACML_DATATYPE_STRING,
                "",
                resource_x509_issuer);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }
    /************ E: Resource *********/

    return 0;
}


int pep_construct_request_action   (xacml_request_t   request,
                                    char *            action_type
                                   )
{
    /************ B: Action *********/
    if (action_type)
    {
        xacml_request_add_action_attribute(
                request,
                XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID,
                XACML_DATATYPE_STRING,
                "",
                action_type);
    }
    /************ E: Action *********/

    return 0;
}



/******************************************************************************
Function:   get_gidlist()

Description:
    Finds the list of gids for user in the group file (/etc/group)
    Returns a list of gid_t which should be freed by calling program.

Parameters:
    username:   the name of the user
    ngroups:    ptr to int which will be filled with the number of gids.
    group_list: ptr to an array of gid_t.

Returns:
    0 on success. 
    -1 on realloc failure
    -2 on getgrent failure
    1 on failure
******************************************************************************/
int get_gidlist(
        const char * username,
        int * ngroups,
        gid_t ** group_list
    )
{
    struct group        * group_info = NULL;
    gid_t               * groups = NULL;
    gid_t               * newgroups = NULL;
    int                   i = 0;

    /* rewind the file pointer to the beginning of the /etc/group file */
    setgrent();

    lcmaps_log_debug(2, "\tlcmaps_get_gidlist(): looping through group file\n");
    *ngroups = 0;
    while ( ( group_info = getgrent() ) )
    {
        char ** pgr_mem = group_info->gr_mem;
        char *  gr_mem = NULL;

        lcmaps_log_debug(4, "\tlcmaps_get_gidlist(): group %s\n", group_info->gr_name);
        while ( (gr_mem = *pgr_mem) )
        {
            lcmaps_log_debug(4, "\tlcmaps_get_gidlist(): \tgroup member %s\n", gr_mem);
            if (strncmp(username, gr_mem, strlen(username))==0)
            {
                lcmaps_log_debug(2, "\tlcmaps_get_gidlist(): \t\tfound group %s for %s\n",
                    group_info->gr_name, username);
                (*ngroups)++;
                newgroups = (gid_t *) realloc(groups, ((*ngroups) * sizeof(gid_t)));
                if (newgroups == NULL)
                {
                    lcmaps_log(1, "lcmaps_get_gidlist(): cannot realloc\n");
                    free(groups);
                    return -1;
                }
                groups=newgroups;
                groups[(*ngroups)-1] = group_info->gr_gid;
            }
            ++pgr_mem;
        }
    }
    if (errno==ENOMEM)
    {
        lcmaps_log(1, "lcmaps_get_gidlist(): Cannot read the group file, %s\n", strerror(errno));
        free(groups);
        groups=NULL;
        /* Close the group file */
        endgrent();
        return -2;
    }
    *group_list=groups;
    lcmaps_log_debug(4,"\tlcmaps_get_gidlist(): %d groups found for %s\n", *ngroups, username);
    for (i = 0; i < *ngroups; i++)
    {
        lcmaps_log_debug(4,"\tlcmaps_get_gidlist(): group nr %d ==> gid_t %d\n", i+1, groups[i]);
    }
    /* Close the group file */
    endgrent();
    return 0;
}



int getOHUIDGIDFired(void)
{
        return oh_uidgid_fired;
}
int getOHSecondaryGIDsFired(void)
{
        return oh_secondarygids_fired;
}
int getOHUsernameFired(void)
{
        return oh_username_fired;
}

void setOHUIDGIDFired(int fired)
{
        oh_uidgid_fired = fired;
}
void setOHSecondaryGIDsFired(int fired)
{
        oh_secondarygids_fired = fired;
}
void setOHUsernameFired(int fired)
{
        oh_username_fired = fired;
}
