/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


#define TRUE   1
#define FALSE  0

/* network functions */
/*
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <unistd.h>
*/


#include "net_common.h"
#include <lcmaps/lcmaps_log.h>



/****************************************************
Function: setup_master_socket
Description:
          Setup Master Socket with all the socket options and such.

Variables:
          int * master_socket : master socket descriptor
          int   port          : port number to listen on
Return:
    0 : Good
   !0 : Not good
****************************************************/
int setup_master_socket (int * master_socket, int port)
{
    int opt = 1;
    struct sockaddr_in address;


    /* create the master socket and check it worked */
    if ((*master_socket = socket(AF_INET,SOCK_STREAM,0)) < 0)
    {
        lcmaps_log (1, "Failure in create master_socket fail\n");
        return 1;
    }

    /* set master socket to allow multiple connections */
    if (setsockopt(*master_socket, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt)) < 0)
    {
        lcmaps_log (1, "Failure in setsockopt\n");
        return 2;
    }

    /* type of socket created */
    address.sin_family      = AF_INET;
    address.sin_addr.s_addr = INADDR_ANY;
    address.sin_port        = htons(port);

    /* bind master socket to port number */
    if (bind(*master_socket, (struct sockaddr *)&address, sizeof(address)) < 0)
    {
        lcmaps_log (1, "Failure in bind\n");
        return 3;
    }


    /* try to specify maximum of  15 pending connections 
     * for  the master socket */
    if (listen(*master_socket, 15) < 0)
    {
        lcmaps_log (1, "Failure in setting socket to listen\n");
        return 4;
    }

    return 0;
}


/****************************************************
Function: accept_new_client_socket
Description:
          Accepts new client socket from a master socket and returns that with the 
          DNS canonical name of the connected remote host

Variables:
          int master_socket   : master socket descriptor
          int * client_socket : return pointer to the client_socket descriptor
          char ** remote_host : DNS canonical name string of the remote host 
                                connecting to the master socket
Return:
    0 : Good
   !0 : Not good
****************************************************/
int accept_new_client_socket (int              master_socket, 
                              int             *client_socket, 
                              struct sockaddr_in *addr,
                              socklen_t       *addr_len,
                              char *          *remote_host)
{
    int new_socket = -1;

    lcmaps_log_debug (5, "Accepting new client connection\n");

    /* Open the new socket as 'new_socket' */
    if ((new_socket = accept(master_socket, (struct sockaddr *) addr, addr_len)) < 0)
    {
        lcmaps_log (3, "Accept of client socket fail!\n");
        return 1;
    }

    /* Get remote_host string from opened connection */
    *remote_host = (char *) malloc (sizeof(char) * 255);
    addr2host (remote_host, 255, (struct sockaddr_in) *addr);

    lcmaps_log_debug (4, "Service: Accepted connection from remote host: %s\n", *remote_host);

    /* add new socket to list of sockets */
    *client_socket = new_socket;

    return 0;
}



/****************************************************
Function: addr2host
Variables:
          char ** hostname         : buffer in which the hostname will be written (requires preallocation)
          int     bufsize          : size of hostname buffer
          struct sockaddr_in  addr : socket address information (for Internet)
Return:
    0 : Good
   !0 : Not good
****************************************************/
int addr2host (char **            hostname,
               int                bufsize,
               struct sockaddr_in addr)
{
    struct hostent * hp = NULL;

    /* Map the IP address to a host name. */
    hp = gethostbyaddr ((char *)&addr.sin_addr, sizeof(struct in_addr), addr.sin_family);
    if (hp)
        strncpy(*hostname, hp->h_name, bufsize);
    else
        strncpy(*hostname, inet_ntoa(addr.sin_addr), bufsize);

    return 0;
}







int create_client_socket (int * client_socket, 
                          const char * server, 
                          int port, 
                          int time_out_milliseconds)
{
    char                               *logstr = "create_client_socket";
    struct addrinfo                     hints;
    struct addrinfo                    *res;
    int                                 rc;
    int                                 mysock = -1;
    char                                portstr[24];

    struct timeval *                    wait_tv = NULL;
    struct timeval                      preset_tv;
    unsigned int                        preset_tvlen = sizeof preset_tv;


    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_socktype = SOCK_STREAM;
    hints.ai_family   = AF_INET;

    /* Get addrinfo */
    snprintf(portstr, 24, "%d", port);
    rc = getaddrinfo(server, &portstr[0], &hints, &res);
    if (rc != 0)
    {
        lcmaps_log (1, "Error: Failed to getaddrinfo (%s, %s, *, *)\n", server, portstr);
        return 1;
    }


    /* Create new socket */
    if ((mysock = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0)
    {
        lcmaps_log (1, "Error: Failed to create socket\n");
        freeaddrinfo( res );
        return 1;
    }

    
    /* Grab timeout setting */
    if (getsockopt (mysock, SOL_SOCKET, SO_RCVTIMEO, (char *)&preset_tv, &preset_tvlen) < 0)
    {
        lcmaps_log (1, "%s: Error: Failed to get the timeout setting\n", logstr);
        freeaddrinfo( res );
        return 1;
    }


    /* Set connection timeout on the socket */
    wait_tv = (struct timeval *) malloc (sizeof (struct timeval));
    wait_tv->tv_sec = (time_out_milliseconds - (time_out_milliseconds % 1000)) / 1000;
    wait_tv->tv_usec = (time_out_milliseconds % 1000) * 1000;
    if (setsockopt (mysock, SOL_SOCKET, SO_RCVTIMEO, (char *)wait_tv, sizeof *wait_tv) < 0)
    {
        lcmaps_log (1, "%s: Error: Failed to set the timeout setting\n", logstr);
        freeaddrinfo( res );
        return 1;
    }
    free (wait_tv);
    wait_tv = NULL;


    /* Connecting socket to host on port with timeout */
    if (connect(mysock, res -> ai_addr, res -> ai_addrlen) < 0)
    {
        lcmaps_log_debug (3, "Failed to connect\n");
        freeaddrinfo( res );
        return 1;
    }
    else
    {
        /* Socket is succesfuly connected */
        setsockopt (mysock, SOL_SOCKET, SO_KEEPALIVE, 0, 0);

        *client_socket = mysock;
        freeaddrinfo( res );
        return 0;
    }

    freeaddrinfo( res );

    /* Failure */
    return 1;
}

