/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


#include "ssl-common.h"
#include <lcmaps/lcmaps_log.h>

#ifdef USE_GRIDSITE
    #include "gridsite.h"
#endif

SSL_CTX * ssl_ctx    = NULL;
char * certCAPath    = NULL;
char * certFilePub   = NULL;
char * certFilePriv  = NULL;
char * certOwner     = NULL;

char * ssl_override_hostname = NULL;


/******************************************************************************
Function:     SSL_recv
Description:
              This function will use the SSL object to readout the SSL transport 
              receive buffer for num amount of bytes and put it to the data buffer.
              Error message handling included
              Note the entire SSL read buffer will be read, until it doesn't fit 
              in the buffer.
Parameters:
Returns:
     < 0  :  Error code from SSL_read()
    == 0  :  Nothing to be done
     > 0  :  Amount of bytes read from the buffer
******************************************************************************/
int SSL_recv (SSL * ssl, void * data, int num)
{
    char * logstr = "SSL_recv";
    int error_code = 0;
    int rc = 0;
    int nread = 0;
    int retry = 0;

    lcmaps_log_debug (5, "%s: entering %s\n", logstr, logstr);

    if (!ssl)
    {
        lcmaps_log (1, "%s: No SSL object presented to SSL_recv()\n", logstr);
        return -1;
    }

    /* clear buffer */
    bzero (data, num);

    /* rc is the return code when rc <= 0, otherwise its the amount of bytes read */

    /*
     * Igor's patch to be more resiliant to long distance connections 
     * which don't fill the buffers fast enough to let this work at once
     *
    rc = SSL_read (ssl, data, num);
    if (rc < 0)
        print_ssl_error_msg (ssl, rc, 1, &error_code); 
    */

    /* Endless retry (this might bite me) */
    while (1) 
    {
        rc = SSL_read (ssl, data, num);
        if (rc < 0) 
        {
            int unrecoverable = print_ssl_error_msg (ssl, rc, 1, &error_code);
            if (unrecoverable)
                break; /* fatal error */
            /* else retry */
        } 
        else 
        {
            break; /* no error */
        }
    } 


    lcmaps_log_debug (5, "%s: Read %d bytes, buffer size is %d.\n", logstr, rc, num);
    return rc;


  /* Out of scope */
    while ((nread < num) && 
           (rc = SSL_read (ssl, (unsigned char *) data + nread, num - nread)) && 
           (retry < 5))
    {
        /* Got bytes */
        if (rc > 0)
        {
            nread += rc;
        }
        else
        {
            if (print_ssl_error_msg (ssl, rc, 1, &error_code))
                break;
            else 
            {
                retry++;
                lcmaps_log_debug (5, "%s: continueing SSL_read(), retry %d\n", logstr, retry);
                usleep (50 * 1000 * retry);
                continue;
            } 
        }
    }
    lcmaps_log_debug (5, "%s: Read %d bytes, buffer size is %d.\n", logstr, nread, num);

    print_ssl_error_msg (ssl, rc, 0, &error_code); 
    if ((error_code == SSL_ERROR_WANT_READ) || (error_code == SSL_ERROR_WANT_WRITE))
    {
        lcmaps_log (3, "%s: Something is still in the transport buffer\n", logstr);
    }

    return nread ? nread : rc;
}



/******************************************************************************
Function:     SSL_server_accept
Description:
              This will take the open socket descriptor and the pre-setup CTX,
              connect the socket with the SSL layer and initiate the SSL_connect() 
              handshake 

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL object pointer
    NULL  : FUBAR
******************************************************************************/
SSL * SSL_server_accept (int opensocket, char * remote_host, SSL_CTX * ctx)
{
    char * logstr = "SSL_server_accept";
    int    rc = 0;
    SSL *  ssl  = NULL;
    BIO *  bio  = NULL;
    int              i     = 0;
    int              depth = 0;
    int              ssl_error_code = 0;
    X509 *           ssl_peer_cert       = NULL;
    STACK_OF(X509) * ssl_peer_cert_chain = NULL;
    char           * dn;


    /* Creating the SSL object out of the CTX */
    if (ctx)
    {
        /* Creating the SSL layer */
        if (!(ssl = SSL_new(ctx)))
        {
            lcmaps_log (1, "%s: Error: Failed to create the SSL layer\n", logstr);
            return NULL;
        }
        else
        {
            lcmaps_log_debug (5, "%s: SSL object created from CTX\n", logstr);
            SSL_clear(ssl);
        }
    } 
    else
    {
        lcmaps_log (1, "%s: Error: can't create SSL handle out of CTX structure\n", logstr);
        return NULL;
    }


    /* Connecting the Socket to the SSL layer */
    if (!(bio = BIO_new_socket (opensocket, BIO_NOCLOSE)))
    {
        lcmaps_log (1, "%s: Error: Failed to tie the socket to a SSL BIO\n", logstr);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        lcmaps_log_debug (5, "%s: BIO created from socket\n", logstr);
    }


    /* Tie the socket as a file descriptor to the SSL layer */
    if (! SSL_set_fd(ssl, opensocket))
    {
        lcmaps_log (1, "%s: Error: SSL_set_fd() failed!\n", logstr);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        lcmaps_log_debug (5, "%s: Passed setting binding fd to SSL\n", logstr);
    }


    /* Set the connection to non-blocking mode */
    /*
    BIO_set_nbio (bio, 1);
     */

    SSL_set_bio(ssl, bio, bio);
    lcmaps_log_debug (5, "%s: BIO set for SSL\n", logstr);

    SSL_set_accept_state(ssl);
    lcmaps_log_debug (5, "%s: Passed SSL set accept state\n", logstr);


    /* SSL accepting */
    if ((rc = SSL_accept(ssl)) <= 0)
    {
        lcmaps_log_debug (3, "%s: Warning: failure in SSL_accept()\n", logstr);
        if (print_ssl_error_msg (ssl, rc, 1, &ssl_error_code))
        {
            SSL_free(ssl);
            return NULL;
        } 
        else
        {
            if ((ssl_error_code == SSL_ERROR_WANT_READ) || (ssl_error_code == SSL_ERROR_WANT_WRITE))
            {
                lcmaps_log (3, "%s: Something is still in the transport buffer\n", logstr);
            }

            /* SSL accepting - retry */
            if ((rc = SSL_accept(ssl)) <= 0)
            {
                lcmaps_log (1, "%s: Error: failure in SSL_accept() (already retried)\n", logstr);
                print_ssl_error_msg (ssl, rc, 1, NULL);
                SSL_free(ssl);
                return NULL;
            } 
        }
    }
    lcmaps_log_debug (5, "Successful SSL_accept()\n");

/***/

    if (!(ssl_peer_cert = SSL_get_peer_certificate(ssl)))
    {
        lcmaps_log (1, "%s: Error: Could not get peer certificate from the SSL handle.\n", logstr);
        SSL_free(ssl);
    }

    if (!(ssl_peer_cert_chain = SSL_get_peer_cert_chain(ssl)))
    {
        lcmaps_log (1, "%s: Error: Could not get peer chain of certificates from the SSL handle\n", logstr);
        SSL_free(ssl);
    }
    else
    {
        /*
         * So we have a chain... but does it contain the peer certificate?
         * ... lets check and add it if not
         */

        depth = sk_X509_num(ssl_peer_cert_chain);
        lcmaps_log_debug (5, "%s: Chain depth = %d\n", logstr, depth);
        sk_X509_insert(ssl_peer_cert_chain, ssl_peer_cert, 0);
        lcmaps_log_debug (5, "%s: Added the peer certificate to the chain\n", logstr);
        depth = sk_X509_num(ssl_peer_cert_chain);
        lcmaps_log_debug (5, "%s: Chain depth = %d\n", logstr, depth);

        for (i = 0; i < depth; i++)
        {
            dn = X509_NAME_oneline(X509_get_subject_name (sk_X509_value(ssl_peer_cert_chain, i)), NULL, 0);

            lcmaps_log_debug (5, "%s: (%d) -> \"%s\" (%s) .\n", 
                               logstr, 
                               i, 
                               dn,
                               x509IsCA(sk_X509_value(ssl_peer_cert_chain, i)) ? "is CA" : "not a CA");

            free( dn );
        }
    }


    /* Connection stuff done, client connected */
    lcmaps_log_debug (4, "%s: SSL Connection accepted and opened\n", logstr);

    return ssl;
}



/******************************************************************************
Function:     SSL_client_connect
Description:
              This will take the open socket descriptor and the pre-setup CTX,
              connect the socket with the SSL layer and initiate the SSL_connect() 
              handshake 

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL object pointer
    NULL  : FUBAR
******************************************************************************/
SSL * SSL_client_connect (int opensocket, SSL_CTX * ctx) 
{
    int    rc = 0;
    SSL *  ssl  = NULL;
    BIO *  bio  = NULL;
    int    retry = 0;


    /* Creating the SSL object out of the CTX */
    if (ctx)
    {
        /* Creating the SSL layer */
        if (!(ssl = SSL_new(ctx)))
        {
            lcmaps_log (1, "%s: Error: Failed to create the SSL layer. OpenSSL reason: %s.\n", 
                        __func__,
                        ERR_reason_error_string(ERR_get_error()));
            return NULL;
        }
        else
        {
            lcmaps_log_debug (5, "%s: SSL object created from CTX\n", __func__);
        }
    }
    else
    {
        lcmaps_log (1, "%s: Error: can't create SSL handle out of CTX structure\n", __func__);
        return NULL;
    }



    /* Connecting the Socket to the SSL layer */
    if (!(bio = BIO_new_socket (opensocket, BIO_NOCLOSE)))
    {
        lcmaps_log (1, "%s: Error: Failed to tie the socket to a SSL BIO\n", __func__);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        lcmaps_log_debug (5, "%s: BIO created from socket\n", __func__);
    }


    /* Set the connection to non-blocking mode */
    /* 
    BIO_set_nbio (bio, 1);
     */

    /* Binding/setting the BIO to the SSL layer */
    SSL_set_bio(ssl, bio, bio);
    lcmaps_log_debug (5, "%s: BIO set for SSL\n", __func__);


    /* The retries are local to the system mostly, hence you can try a lot per connected client */
    if ((rc = SSL_connect(ssl)) <= 0)
    {
        lcmaps_log (1, "%s: Error connecting SSL client object\n", __func__);
        if (print_ssl_error_msg (ssl, rc, 1, NULL))
        {
            SSL_free(ssl);
            return NULL;
        } 
        else
        {
            while (retry < 222)
            {
                if ((rc = SSL_connect(ssl)) <= 0)
                {
                    if (print_ssl_error_msg (ssl, rc, 0, NULL))
                    {
                        break; 
                    } 
                    else
                    {
                        /* According to the SSL error, you should try again. Introducing and I/O micro sleep */
                        retry++;
                        continue;
                    }
                }
                else
                {
                    /* Success */
                    lcmaps_log_debug (5, "%s: SSL Connection opened\n", __func__);
                    return ssl;
                }
            }

            /* Non-succesfull / failure condition */
            /* If here, then the SSL connection isn't meant to be, so bailing out */
            print_ssl_error_msg (ssl, rc, 1, NULL);
            SSL_free(ssl);
            return NULL;
        }
    }
    else
    {
        /* Success */
        lcmaps_log_debug (5, "%s: SSL Connection opened\n", __func__);
        return ssl;
    }
}


/******************************************************************************
Function:     SSL_connect_shutdown
Description:
              The SSL object will be shutdown fully

Parameters:
Returns:
        0 : Perfect shutdown
        1 : FUBAR
******************************************************************************/
int SSL_connect_shutdown (SSL * ssl)
{
    int retry = 0;
    int rc = 0;

    if (!ssl)
    {
        lcmaps_log_debug (5, "Shutting down SSL; nothing to shutdown though... empty SSL pointer presented.\n");
        return 1;
    }
    else
    {
        lcmaps_log_debug (4, "Shutting down SSL\n");
        while (retry < 10)
        {
          switch (rc = SSL_shutdown (ssl))
          {
            /* First SSL_shutdown must return 0 */
            /* Second SSL_shutdown must return 1 */
            case 0 :
                if (retry == 0)
                    lcmaps_log_debug (5, "First shutdown succeeded, close notify send ok\n");
                else
                    lcmaps_log_debug (2, "Retrying shutdown procedure (try: %d)\n", retry + 1);

                retry++;

                /* Perform microsleep to cut the system some slack to catch up */
                usleep (500 + 500 * retry); 
                continue;
            case 1 :
                lcmaps_log_debug (1, "Succesfull shutdown #1\n");

                lcmaps_log_debug (5, "Freeing SSL BIO's\n");
                BIO_free( SSL_get_rbio( ssl ) );

                SSL_CTX_sess_set_remove_cb(SSL_get_SSL_CTX( ssl ), NULL); 
                SSL_CTX_free( SSL_get_SSL_CTX( ssl ) );

                lcmaps_log_debug (1, "Succesfull shutdown #2\n");
                return 0;
            default :
                lcmaps_log (3, "Shutdown failed completely, retry count is: %d\n", retry + 1);
                print_ssl_error_msg (ssl, rc, 1, NULL);

                lcmaps_log_debug (5, "Freeing SSL BIO's\n");
                BIO_free( SSL_get_rbio( ssl ) );

                SSL_CTX_sess_set_remove_cb(SSL_get_SSL_CTX( ssl ), NULL); 
                SSL_CTX_free( SSL_get_SSL_CTX( ssl ) );

                return 1;
           }
        }
    }

    /* Shouldn't get here */
    return 1;
}




/******************************************************************************
Function:   setup_client_ctx
Description:
    Setting up the SSL client side context with the right settings

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL_CTX context pointer
    NULL  : FUBAR
******************************************************************************/
SSL_CTX *setup_client_ctx(char * cafile, 
                          char * cadir, 
                          char * clientcert, 
                          char * clientkey)
{
    SSL_CTX * ctx = NULL;

    uid_t stored_real_uid = -1;
    uid_t stored_eff_uid  = -1;


    lcmaps_log_debug (5, "Entering: %s\n", __func__);
    ctx = SSL_CTX_new(SSLv23_client_method());
    if (!ctx)
    {
        lcmaps_log (1, "%s: Error: No SSL context was created: %s.\n", 
                    __func__, 
                    ERR_reason_error_string(ERR_get_error()));
        return NULL;
    }


    if (SSL_CTX_load_verify_locations(ctx, cafile, cadir) != 1)
    {
        lcmaps_log (1, "%s: Error loading CA file and/or directory: file: %s dir: %s. OpenSSL reason: %s.\n", 
                    __func__, 
                    cafile ? cafile : "n/a", 
                    cadir ? cadir : "n/a",
                    ERR_reason_error_string(ERR_get_error()));
        SSL_CTX_free (ctx);
        return NULL;
    }


    /* Change account, by default to bypass NFS root squash problems */
    changeEffectiveToRealUid (&stored_real_uid, &stored_eff_uid);

    if (SSL_CTX_use_certificate_chain_file(ctx, clientcert) != 1) 
/*    if (SSL_CTX_use_certificate_file(ctx, clientcert, SSL_FILETYPE_PEM) != 1) */
/*         lcmaps_log (1, "%s: Error loading certificate from file", __func__); */
    {
        lcmaps_log (1, "%s: Error loading certificate chain from file: '%s'. OpenSSL reason: %s.\n", 
                    __func__, 
                    clientcert ? clientcert : "n/a",
                    ERR_reason_error_string(ERR_get_error()));
        SSL_CTX_free (ctx);
        return NULL;
    }

    if (SSL_CTX_use_PrivateKey_file(ctx, clientkey, SSL_FILETYPE_PEM) != 1)
    {
        lcmaps_log (1, "%s: Error loading private key from file: '%s'. OpenSSL reason: %s.\n", 
                    __func__, 
                    clientkey,
                    ERR_reason_error_string(ERR_get_error()));
        SSL_CTX_free (ctx);
        return NULL;
    }

    /* Restore to the previous settings */
    restoreEffectiveToRealUid (stored_real_uid, stored_eff_uid);


    /*
     * Go into verify_callback after default SSL checks 
     * Perform signature checks (peer must present certificate and must verify it)
     */
#ifndef DISABLE_PROXY_SUPPORT
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, scas_verify_callback);

  #if OPENSSL_VERSION_NUMBER < 0x00908000L
    X509_STORE_set_flags(ctx->cert_store, X509_V_FLAG_CRL_CHECK |
                                          X509_V_FLAG_CRL_CHECK_ALL );
    ctx->cert_store->check_issued = grid_X509_check_issued_wrapper;
  #else
    X509_STORE_set_flags(ctx->cert_store, X509_V_FLAG_CRL_CHECK |
                                          X509_V_FLAG_CRL_CHECK_ALL |
                                          X509_V_FLAG_ALLOW_PROXY_CERTS );
  #endif /* OPENSSL_VERSION_NUMBER < 0x00908000L */
#endif /* DISABLE_PROXY_SUPPORT */


    /* Set max Chain depth */
    SSL_CTX_set_verify_depth(ctx, MAXCHAINDEPTH);
    SSL_CTX_set_options(ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);
    if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST) != 1)
    {
        lcmaps_log (1, "%s: Error setting cipher list (no valid ciphers)\n", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    return ctx;
}


/******************************************************************************
Function:   setup_server_ctx
Description:
    Setting up the SSL server side context with the right settings

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL_CTX context pointer    
    NULL  : FUBAR
******************************************************************************/
SSL_CTX *setup_server_ctx(char * cafile,
                          char * cadir,
                          char * servercert,
                          char * serverkey)
{
    char * logstr = "setup_server_ctx";
    SSL_CTX *ctx;

    lcmaps_log_debug (5, "Entering: %s\n", logstr);
    /* ctx = SSL_CTX_new(SSLv23_method()); */

    lcmaps_log_debug (5, "%s: creating context\n", logstr);
    ctx = SSL_CTX_new(SSLv23_server_method());

    if (!ctx)
    {
        lcmaps_log (1, "%s: Error: No SSL context was created\n", logstr);
        return NULL;
    }
    else
    {
        lcmaps_log_debug (5, "%s: Context created\n", logstr);
    }


    if (SSL_CTX_use_PrivateKey_file(ctx, serverkey, SSL_FILETYPE_PEM) != 1)
    {
        lcmaps_log (1, "%s: Error loading private key from file '%s'\n", logstr, serverkey);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    { 
        lcmaps_log_debug (5, "%s: Private Key added to ctx\n", logstr);
    }


    /* 
     * Don't set the default OpenSSL verify paths with SSL_CTX_set_default_verify_paths(ctx).
     * Only the configured (Grid) trust anchors are trusted
     */
    if (SSL_CTX_load_verify_locations(ctx, cafile, cadir) != 1)
    {
        lcmaps_log (1, "%s: Error loading CA file and/or directory '%s'\n", logstr, cafile);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    { 
        lcmaps_log_debug (5, "%s: CA directory added to ctx\n", logstr);
    }

    /* if (SSL_CTX_use_certificate_file(ctx, servercert, SSL_FILETYPE_PEM) != 1) */
    if (SSL_CTX_use_certificate_chain_file(ctx, servercert) != 1) 
    {
        lcmaps_log (1, "%s: Error loading certificate from file '%s'\n", logstr, servercert);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    { 
        lcmaps_log_debug (5, "%s: Public certificate chain added to ctx\n", logstr);
    }


    /*
     * Go into verify_callback after default SSL checks 
     * Perform signature checks (peer must present certificate, must verify peer
     */


#ifndef DISABLE_PROXY_SUPPORT
    SSL_CTX_set_purpose(ctx, X509_PURPOSE_ANY);
/*  X509_STORE_CTX_set_purpose( verify_ctx, X509_PURPOSE_SSL_CLIENT ); */

    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, scas_verify_callback);
/*  SSL_CTX_set_cert_verify_callback(ctx, scas_gsi_callback_X509_verify_cert, NULL); */
 
  #if OPENSSL_VERSION_NUMBER < 0x00908000L
    X509_STORE_set_flags(SSL_CTX_get_cert_store(ctx), X509_V_FLAG_CRL_CHECK |
                                                      X509_V_FLAG_CRL_CHECK_ALL );
    ctx->cert_store->check_issued = grid_X509_check_issued_wrapper;
  #else
    X509_STORE_set_flags(SSL_CTX_get_cert_store(ctx), X509_V_FLAG_CRL_CHECK |
                                                      X509_V_FLAG_CRL_CHECK_ALL |
                                                      X509_V_FLAG_ALLOW_PROXY_CERTS );
  #endif /* OPENSSL_VERSION_NUMBER < 0x00908000L */

#if 0 /* Should re-check this flag's placement. Normally found in or before the verify function set to the individual certificate from an SSL handshake. */
   #if OPENSSL_VERSION_NUMBER >= 0x00908000L
     cert->ex_flags |= EXFLAG_PROXY;
   #endif /* OPENSSL_VERSION_NUMBER >= 0x00908000L */
#endif

    /* Max depth */
    SSL_CTX_set_verify_depth(ctx, MAXCHAINDEPTH);
#endif /* DISABLE_PROXY_SUPPORT */



/**************************************************************************************/


    /* All (SSLv3, TLS) but not SSLv2 */
    SSL_CTX_set_options(ctx, SSL_OP_ALL | SSL_OP_NO_SSLv2);

    if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST) != 1)
    {
        lcmaps_log (1, "%s: Error setting cipher list (no valid ciphers)\n", logstr);
        SSL_CTX_free (ctx);
        return NULL;
    }

    /* Does the Private key match the certificate? */
    if (SSL_CTX_check_private_key(ctx) != 1)
    {
        lcmaps_log (1, "%s: Error matching private key\n", logstr);
        SSL_CTX_free (ctx);
        return NULL;
    }

    return ctx;
}




/******************************************************************************
Function:   x509IsCA
Description:
    Tests if the X509 * cert is a CA certificate or not
    => Example from GridSite
Parameters:
    A X509 pointer
Returns:
    0      : Not a CA cert
    1      : This is a CA cert
******************************************************************************/
int x509IsCA (X509 *cert)
{

/*
#define NEW_CA_SEARCH_METHOD
*/
#ifndef NEW_CA_SEARCH_METHOD
    int purpose_id;

    purpose_id = X509_PURPOSE_get_by_sname("sslclient");

    /* final argument to X509_check_purpose() is whether to check for CAness */

    if (X509_check_purpose(cert, purpose_id + X509_PURPOSE_MIN, 1))
        return 1;
    else return 0;
#else
    int                                 element_index = -1;
    int                                 critical;
    BASIC_CONSTRAINTS *                 x509v3_bc = NULL;

    if((x509v3_bc = X509_get_ext_d2i(cert,
                                     NID_basic_constraints,
                                     &critical,
                                     &element_index)) && x509v3_bc->ca)
    {
        /* Free allocated space */
        BASIC_CONSTRAINTS_free(x509v3_bc);

        /* CA cert found */ 
        return 1;
    }

    /* Not a CA cert */
    return 0;
#endif /* NEW_CA_SEARCH_METHOD */
}




/******************************************************************************
Function:       x509_chain_to_dn()
Description:    Get the base DN of the X509 proxy cert
                Watch out! This function changes the subject DN in the X509 structure!
Parameters:
                px509: x509 certificate
                certstack: certificate chain
Returns:
        dn (which should be freed)
******************************************************************************/
char * x509_chain_to_dn(
        X509 * px509, STACK_OF(X509) *chain)
{
    char *  logstr = "x509_chain_to_dn";
    char *  dn = NULL;    
/*
    STACK_OF (X509) * chain = (STACK_OF(X509) *) sk_X509_dup(certstack);
*/


    if (chain)    
    {        
        int amount_of_CAs = 0;        
        int i = 0;        
        int depth = sk_X509_num(chain);        
        X509 * cert = NULL;        

        lcmaps_log_debug (5, "%s: Chain depth = %d\n", logstr, depth);

        /* How many CA certs are there in the chain? */
        for (i = 0; i < depth; i++)
        {
            dn = X509_NAME_oneline(X509_get_subject_name (sk_X509_value(chain, i)), NULL, 0);
            lcmaps_log_debug (5, "%s: DN here is: %s\n", logstr, dn );
            free( dn );

            if (x509IsCA(sk_X509_value(chain, i)))
                amount_of_CAs++;
        }

        /* Getting the user cert (filtering the CAs and such) */
        i = depth - (amount_of_CAs + 1);
        cert = sk_X509_value(chain, i);
        if (cert)
        {
            dn = X509_NAME_oneline(X509_get_subject_name (cert), NULL, 0);
            lcmaps_log_debug (5, "%s: Extracted DN: %s\n", logstr, dn);
        }
        else
        {
            lcmaps_log_debug (1, "%s: Couldn't extract the requested user cert from the stack, thus can't convert to DN\n", logstr);
        }
        /* clean up */
        /* sk_X509_free (chain); */
    }
    else
    {
        if (px509)
        {
            dn = X509_NAME_oneline(X509_get_subject_name (px509), NULL, 0);
            lcmaps_log_debug (5, "%s: Extracted DN: %s\n", logstr, dn);
        }
        else
        {
            lcmaps_log_debug (1, "%s: Empty certificate stack and empty single certificate, can't convert to DN\n", logstr);
        }
    }

    return dn;
}



/******************************************************************************
Function:       init_OpenSSL()
Description:    Do the SSL library initialization
Parameters:     void
Returns:
     0 : all is ok
     1 : failure reported must not continue with the SSL library
******************************************************************************/
int init_OpenSSL(void)
{
    if (!SSL_library_init())
    {
        lcmaps_log (1, "%s: ** OpenSSL initialization failed!\n", __func__);
        return 1;
    }
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();

    lcmaps_log_debug (5, "%s: OpenSSL initialized\n", __func__);

    return 0;
}


#ifndef DISABLE_PROXY_SUPPORT
/******************************************************************************
Function:       scas_verify_callback()
Description:    Execute the OpenSSL verify callback
                The ok will be the result of the default verification function
Parameters:     ok, X509_STORE_CTX *
Returns:        int
******************************************************************************/
int scas_verify_callback(int ok, X509_STORE_CTX *store_ctx)
{
    unsigned long   errnum   = X509_STORE_CTX_get_error(store_ctx);
    int             errdepth = X509_STORE_CTX_get_error_depth(store_ctx);
    const char *    logstr = "verify_callback";
    char            subject[256];
    char            issuer[256];

    X509 * curr_cert = X509_STORE_CTX_get_current_cert(store_ctx);

    lcmaps_log_debug (3, "%s: - Re-Verify certificate at depth: %i\n", logstr, errdepth);
    X509_NAME_oneline(X509_get_issuer_name(curr_cert), issuer, 256);
    lcmaps_log_debug (3, "%s:  issuer   = %s\n", logstr, issuer);
    X509_NAME_oneline(X509_get_subject_name(curr_cert), subject, 256);
    lcmaps_log_debug (3, "%s:  subject  = %s\n", logstr, subject);
    lcmaps_log_debug (3, "%s:  errnum %d: %s\n", logstr, (int) errnum, X509_verify_cert_error_string(errnum));


    if (ok != 1)
    {
#if OPENSSL_VERSION_NUMBER < 0x00908000L
        if (errnum == X509_V_ERR_INVALID_CA)
        {
            lcmaps_log_debug (3,  "%s: Override 'Invalid CA' error.\n", logstr);
            ok=1;
        }
#endif
        if (errnum == X509_V_ERR_UNABLE_TO_GET_CRL)
        {
            lcmaps_log_debug (3,  "%s: Override 'Unable to get CRL' error\n", logstr);
            ok=1;
        }
        if (errnum == X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION)
        {
            lcmaps_log_debug (3,  "%s: Override 'Unhandled Critical Extensions' error. Continue in grid_X509_knownCriticalExts()\n", logstr);
            errnum = grid_X509_knownCriticalExts(curr_cert);
            store_ctx -> error = errnum;
            if (errnum == X509_V_OK)
            {
                lcmaps_log_debug (3,  "%s: \tgrid_X509_knownCriticalExts() returned X509_V_OK.\n", logstr);
                ok=1;
            }
            else
            {
                lcmaps_log_debug (3,  "%s: \tfailure: grid_X509_knownCriticalExts() returned: %d\n", logstr, errnum);
            }
        }

        if (store_ctx->error == X509_V_ERR_INVALID_PURPOSE)
        {
            lcmaps_log_debug (3,  "%s: Override 'Unsupported certificate purpose' error\n", logstr);
            ok=1;
        }
#if 0
        if (ctx->error == X509_V_ERR_CERT_HAS_EXPIRED) ok=1;
        /* since we are just checking the certificates, it is
         * ok if they are self signed. But we should still warn
         * the user.
         */
        if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
        /* Continue after extension errors too */
        if (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED) ok=1;
        if (ctx->error == X509_V_ERR_INVALID_PURPOSE) ok=1;
        if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
        if (ctx->error == X509_V_ERR_CRL_HAS_EXPIRED) ok=1;
        if (ctx->error == X509_V_ERR_CRL_NOT_YET_VALID) ok=1;
#endif

        if (ok != 1)
        {
            lcmaps_log (1, "%s: Verification failure at depth %d\n", logstr, errdepth);
            lcmaps_log_debug (3, "%s: \tFailure still applies, no override performed for this error\n", logstr);
            lcmaps_log (2, "%s:  \tissuer   = %s\n", logstr, issuer);
            lcmaps_log (2, "%s:  \tsubject  = %s\n", logstr, subject);
        }
    }

    /*
     * We've now got the last certificate - the identity being used for
     * this connection. At this point we check the whole chain for valid
     * CAs or, failing that, GSI-proxy validity using grid_verifyProxy
     */
    if ( (errdepth == 0) && (ok == 1) )
    {
        STACK_OF(X509) *certstack = X509_STORE_CTX_get1_chain(store_ctx);

        lcmaps_log_debug (5, "%s: Final certificate in the stack reached, starting full chain test in grid_verifyProxy(certstack)\n", logstr);
        errnum = grid_verifyProxy(certstack);
        store_ctx->error = errnum;
        if (errnum == X509_V_OK) ok=1;

        /* Revived this pop_free - Brian: Nov 12 2010 */
         /* Originally, this memory leak was required because of a later double-free */
        sk_X509_pop_free( certstack, X509_free );
        certstack = NULL;
    }
    return ok;
}


/******************************************************************************
Function:       grid_X509_knownCriticalExts
Description:    Check if the Critical Extention known to proxy certificates
                are set and understood.
Parameters:     X509 * cert
Returns:        unsigned long : X509_V_OK (good), X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION (bad)
******************************************************************************/
#define PROXYCERTINFO_OID      "1.3.6.1.5.5.7.1.14"
#define OLD_PROXYCERTINFO_OID  "1.3.6.1.4.1.3536.1.222"

unsigned long grid_X509_knownCriticalExts(X509 *cert)
{
   int  i = 0;
   char critical[80];
   X509_EXTENSION *ex;
   char            data[256];
   const char *    logstr = "grid_X509_knownCriticalExts";


   lcmaps_log_debug (5, "Entering %s\n", logstr);
   for (i = 0; i < X509_get_ext_count(cert); ++i)
   {
        X509_NAME_oneline(X509_get_issuer_name(cert), data, 256);
        lcmaps_log (5, "%s:  issuer  = %s\n", logstr, data);
        X509_NAME_oneline(X509_get_subject_name(cert), data, 256);
        lcmaps_log (5, "%s:  subject = %s\n", logstr, data);
        ex = X509_get_ext(cert, i);

        if (X509_EXTENSION_get_critical(ex) &&
                                 !X509_supported_extension(ex))
        {
            OBJ_obj2txt(critical, sizeof(critical), X509_EXTENSION_get_object(ex), 1);

            lcmaps_log (5, "%s: Critical extension found: %s", logstr, critical);
            if (strcmp(critical, PROXYCERTINFO_OID) == 0)     return X509_V_OK;
            if (strcmp(critical, OLD_PROXYCERTINFO_OID) == 0) return X509_V_OK;

            return X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION;
        }
   }
   return X509_V_OK;
}


/******************************************************************************
Function:   grid_verifyProxy
Description:
    Tries to verify the proxies in the certstack
******************************************************************************/
unsigned long grid_verifyProxy( STACK_OF(X509) *certstack )
{
    int      i = 0;
    X509    *cert = NULL;
    time_t   now = time((time_t *)NULL);
    size_t   len = 0;             /* Lengths of issuer and cert DN */
    size_t   len2 = 0;            /* Lengths of issuer and cert DN */
    int      prevIsLimited = 0;   /* previous cert was proxy and limited */
    char     cert_DN   [256];     /* Pointer to current-certificate-in-certstack's DN */
    char     issuer_DN [256];     /* Pointer to issuer-of-current-cert-in-certstack's DN */
    char    *proxy_part_DN = NULL;
    int      depth = sk_X509_num (certstack);
    int      amount_of_CAs = 0;
    const char * logstr = "grid_verifyProxy";

    lcmaps_log_debug (5, "%s: --- Welcome to the grid_verifyProxy function ---\n", logstr);

    if (depth == 0)
    {
        lcmaps_log (1, "%s: Error: Empty certificate chain presented!\n", logstr);
        return X509_V_ERR_APPLICATION_VERIFICATION;
    }


    // And there was (current) time...
    time(&now);

    // How many CA certs are there in the certstack?
    for (i = 0; i < depth; i++)
    {
        if (x509IsCA(sk_X509_value(certstack, i)))
            amount_of_CAs++;
    }


    lcmaps_log_debug (5, "%s: #CA's = %d , depth = %d\n", logstr, amount_of_CAs, depth);

    /* Need to change */
    if ((amount_of_CAs + 1) > depth)
    {
        if ((depth - amount_of_CAs) > 0)
        {
            lcmaps_log (1, "%s: No proxy certificate in certificate stack to check.\n", logstr);
            return X509_V_OK;
        }
        else
        {
            lcmaps_log (1, "%s: No personal certificate (neither proxy or user certificate) found in the certficiate stack.", logstr);
            return X509_V_ERR_APPLICATION_VERIFICATION;
        }
    }


    /* 
     * Changed this value to start checking the proxy and such and
     * to skip the CA and the user_cert
     */

    for (i = depth - (amount_of_CAs + 2); i >= 0; i--)
    {
        /* Check for X509 certificate and point to it with 'cert' */
        if ( (cert = sk_X509_value(certstack, i)) != NULL )
        {
            X509_NAME_oneline(X509_get_subject_name(cert), cert_DN,   256);
            X509_NAME_oneline(X509_get_issuer_name(cert),  issuer_DN, 256);
            len       = strlen( cert_DN );
            len2      = strlen( issuer_DN );

            lcmaps_log_debug (5, "%s: Proxy to verify:\n", logstr );
            lcmaps_log_debug (5, "%s:   Issuer DN: %s\n", logstr, issuer_DN );
            lcmaps_log_debug (5, "%s:   DN:        %s\n", logstr, cert_DN );

            if (now < grid_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notBefore(cert)),0))
            {
                lcmaps_log (1, "%s: Proxy certificate is not yet valid.\n", logstr);
                return X509_V_ERR_CERT_NOT_YET_VALID;
            }

            if (now > grid_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notAfter(cert)),0))
            {
                lcmaps_log (1, "%s: Proxy certificate expired.\n", logstr);
                return X509_V_ERR_CERT_HAS_EXPIRED;
            }

            /* User not allowed to sign shortened DN */
            /* Note: Only true for GT2 proxies */
            if (len2 > len)
            {
                lcmaps_log (1, "%s: It is not allowed to sign a shorthened DN.\n", logstr);
                return X509_V_ERR_INVALID_CA;
            }

            /* Proxy subject must begin with issuer. */
            if (strncmp(cert_DN, issuer_DN, len2) != 0)
            {
                lcmaps_log (1, "%s: Proxy subject must begin with the issuer.\n", logstr);
                return X509_V_ERR_INVALID_CA;
            }

            /* Set pointer to end of base DN in cert_DN */
            proxy_part_DN = &cert_DN[len2];

            /* First attempt at support for Old and New style GSI
               proxies: /CN=anything is ok for now */
            if (strncmp(proxy_part_DN, "/CN=", 4) != 0)
            {
                lcmaps_log (1, "%s: Could not find a /CN= structure in the DN, thus it is not a proxy.\n", logstr);
                return X509_V_ERR_INVALID_CA;
            }


            if ((strncmp(proxy_part_DN, "/CN=limited proxy", 17) == 0) && (i > 0))
            {
                prevIsLimited = 1;
                lcmaps_log (3, "%s:  Found limited proxy.\n", logstr);
            }
            else
            {
                if (prevIsLimited)
                {
                    lcmaps_log (1, "%s: Proxy chain integrity error. Previous proxy in chain was limited, but this one is a regular proxy.\n", logstr);
                    return X509_V_ERR_INVALID_CA;
                }
            }
            lcmaps_log_debug (5, "%s:   Proxy is valid\n", logstr);
        }
    }

    return X509_V_OK;
}

/**
 * Note that timegm() is non-standard. Linux manpage advices the following
 * substition instead.
 */
time_t my_timegm(struct tm *tm)
{
   time_t ret;
   char *tz;

   tz = getenv("TZ");
   setenv("TZ", "", 1);
   tzset();
   ret = mktime(tm);
   if (tz)
       setenv("TZ", tz, 1);
   else
       unsetenv("TZ");
   tzset();

   return ret;
}



/**
 *  (Use ASN1_STRING_data() to convert ASN1_GENERALIZEDTIME to char * if
 *   necessary)
 */

time_t grid_asn1TimeToTimeT(char *asn1time, size_t len)
{
   char   zone;
   struct tm time_tm;

   if (len == 0) len = strlen(asn1time);

   if ((len != 13) && (len != 15)) return 0; /* dont understand */

   if ((len == 13) &&
       ((sscanf(asn1time, "%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */

   if ((len == 15) &&
       ((sscanf(asn1time, "20%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */

   /* time format fixups */

   if (time_tm.tm_year < 90) time_tm.tm_year += 100;
   --(time_tm.tm_mon);

   return my_timegm(&time_tm);
}


#if OPENSSL_VERSION_NUMBER < 0x00908000L
int grid_X509_check_issued_wrapper(X509_STORE_CTX *ctx,X509 *x,X509 *issuer)
/* We change the default callback to use our wrapper and discard errors
   due to GSI proxy chains (ie where users certs act as CAs) */
{
    int ret;
    const char * logstr = "grid_X509_check_issued_wrapper";
    char data[256];

    lcmaps_log_debug (5, "Starting %s\n", logstr);

    X509_NAME_oneline(X509_get_issuer_name(issuer), data, 256);
    lcmaps_log_debug (3, "%s:  issuer   = %s\n", logstr, data);
    X509_NAME_oneline(X509_get_subject_name(x), data, 256);
    lcmaps_log_debug (3, "%s:  subject  = %s\n", logstr, data);

    ret = X509_check_issued(issuer, x);
    if (ret == X509_V_OK)
    {
        lcmaps_log_debug (3, "%s: \tX509_check_issued() == ok\n", logstr);
        return 1;
    }

    /* Non self-signed certs without signing are ok if they passed
           the other checks inside X509_check_issued. Is this enough? */
    if ((ret == X509_V_ERR_KEYUSAGE_NO_CERTSIGN) &&
        (X509_NAME_cmp(X509_get_subject_name(issuer),
                       X509_get_subject_name(x)) != 0))
    {
        lcmaps_log_debug (3, "%s: X509_V_ERR_KEYUSAGE_NO_CERTSIGN && non-self-signed detected.\n", logstr);
        lcmaps_log_debug (3, "%s: \tThese are ok if they pass the other checks inside X509_check_issued() (Is this enough?)\n", logstr);
        return 1;
    }

    lcmaps_log_debug (3, "%s: Not a key usage failure, something else\n", logstr);

    /* If we haven't asked for issuer errors don't set ctx */
    if (!(ctx->flags & X509_V_FLAG_CB_ISSUER_CHECK))
    {
        lcmaps_log_debug (3, "%s: not asked for issuer errors to be set, returning with X509_V_OK\n", logstr);
        return X509_V_OK;
    }

    ctx->error = ret;
    ctx->current_cert = x;
    ctx->current_issuer = issuer;
    ret = ctx->verify_cb(0, ctx);
    lcmaps_log_debug (3, "%s: ctx->verify_cb(0, ctx) == %d\n", logstr, ret);
    return ret;
}
#endif /* OPENSSL_VERSION_NUMBER < 0x00908000L */

#endif /* DISABLE_PROXY_SUPPORT */


/******************************************************************************
Function:       post_connection_check
Description:    Post Connection Checking (only for client)
                DNS (or final /CN=<host>) 
                If SubjectAltNames exist, then they must match the host. If they don't exist, then look at the final CN field in the subject.
                All else is a failure
Parameters:     SSL *, char * host
Returns:        long
******************************************************************************/
long post_connection_check(SSL *ssl, const char *host)
{
    X509      *      peer_cert = NULL;
    X509_NAME *      subj;
    char             datad[256];
    int              extcount;
    int              ok = 0;
    int              found_DNS_entry = 0;
    const char *     logstr = "post_connection_check";

    lcmaps_log_debug (5, "%s: entering %s\n", logstr, logstr);

    if (!host)
    {
        lcmaps_log (1, "%s: error: No hostname supplied to check\n", logstr);
        goto err_occured;
    }

    if (!(peer_cert = SSL_get_peer_certificate(ssl)) || !host)
    {
        lcmaps_log (1, "%s: error: Could not get peer certificate from the SSL handle for host: %s\n",
                     logstr,
                     host);
        goto err_occured;
    }
  

  #if OPENSSL_VERSION_NUMBER >= 0x00908000L
    peer_cert->ex_flags |= EXFLAG_PROXY;
  #endif /* OPENSSL_VERSION_NUMBER >= 0x00908000L */ 

    /* Compare the subjectAltName DNS value with the host value */
    if ((extcount = X509_get_ext_count(peer_cert)) > 0)
    {
        int i;

        for (i = 0; i < extcount; i++)
        {
            X509_EXTENSION    *ext;
            int NID_from_ext = NID_undef; /* Initialize with undefined NID (Numerical ID of a type of ASN1 object) */

            ext = X509_get_ext(peer_cert, i);
            NID_from_ext = OBJ_obj2nid(X509_EXTENSION_get_object(ext));

            if (NID_from_ext == NID_subject_alt_name)
            {
                int                  j;
                unsigned char        *data;
                STACK_OF(CONF_VALUE) *val;
                CONF_VALUE           *nval;
                X509V3_EXT_METHOD    *meth;
                void                 *ext_str = NULL;

                if (!(meth = X509V3_EXT_get(ext)))
                    break;
                data = ext->value->data;

#if (OPENSSL_VERSION_NUMBER > 0x00907000L)
                if (meth->it)
                    ext_str = ASN1_item_d2i(NULL, &data, ext->value->length, ASN1_ITEM_ptr(meth->it));
                else
                    ext_str = meth->d2i(NULL, &data, ext->value->length);
#else
                ext_str = meth->d2i(NULL, &data, ext->value->length);
#endif
                val = meth->i2v(meth, ext_str, NULL);
                for (j = 0;  j < sk_CONF_VALUE_num(val);  j++)
                {
                    nval = sk_CONF_VALUE_value(val, j);
                    if (!strcmp(nval->name, "DNS") && !strcasecmp(nval->value, host))
                    {
                        lcmaps_log_debug (3, "%s: subjectAltName DNS entry matches host\n", logstr);
                        ok = 1;
                        break;
                    }
                    lcmaps_log_debug (5, "%s: found name: %s value: %s\n", logstr, nval->name, nval->value);
                    if (!strcmp(nval->name, "DNS"))
                    {
                        found_DNS_entry = 1;
                    }
                }

                sk_CONF_VALUE_pop_free( val, X509V3_conf_free );
                if (meth->it)
                    ASN1_item_free((ASN1_VALUE*)ext_str, meth->it);
                else
                    meth->ext_free(ext_str);
                ext_str = NULL;

            }
            if (ok)
                break;
        }
    }


    if (found_DNS_entry && !ok)
    {
        lcmaps_log (1, "%s: Error: found one or more DNS entries in the subjectAltName but none match the host '%s'.\n", logstr, host);
        goto err_occured;
    }


    /* If the subjectAltName DNS value doesn't match the host value, 
     * resort back to the X509v1 solution which is to compare the host value 
     * with the final CN value in the cert's subject
     */
    if (!ok && (subj = X509_get_subject_name(peer_cert)) &&
        X509_NAME_get_text_by_NID(subj, NID_commonName, datad, 256) > 0)
    {
        lcmaps_log_debug (5, "%s: No subjectAltName:DNS entry found, trying final CN field in subject to match contacted host.\n", logstr);
        if (strcasecmp(datad, host) != 0)
        {
            lcmaps_log (1, "%s: Error: final CN field in subject '%s' and host '%s' do not match!\n", logstr, datad, host);
            goto err_occured;
        }
        else
        {
            lcmaps_log_debug (5, "%s: final CN field in subject matches host\n", logstr);
            ok = 1;
        }
    }


    /* Clean up */
    if (peer_cert)
    {
        X509_free(peer_cert);
        peer_cert = NULL;
    }

    if (ok)
        return X509_V_OK;
    else
        return X509_V_ERR_APPLICATION_VERIFICATION;

err_occured:
    if (peer_cert)
    {
        X509_free(peer_cert);
        peer_cert = NULL;
    }
    return X509_V_ERR_APPLICATION_VERIFICATION;
}



/******************************************************************************
Function:       seed_prng
Description:    seed prng with /dev/urandom
Parameters:     void
Returns:        void
******************************************************************************/
void seed_prng(void)
{
     RAND_load_file("/dev/urandom", 1024);
}




/****************************************************
Function: print_ssl_error_msg

Variables:
    SSL * ssl   :   pointer to SSL structure (for this connection)
    int ret     :   return code of the previously SSL command
    printmsg    :   0 == don't print the messages, 1 =< print the message

Return:
    0 : Non lethal failure, should consider a retry of the SSL interaction
    1 : Lethal failure, recoverable situation
****************************************************/
int print_ssl_error_msg (SSL * ssl, int ret, int printmsg, int * error_code )
{
    long rc = 0;
    int unrecoverable = 0;

    if (ssl)
    {
        rc = SSL_get_error (ssl, ret);
    }
    else
    {
        rc = ERR_get_error();
    }


    switch (rc)
    {
        case SSL_ERROR_NONE :
        {
            /* No error */
            if (printmsg)
                lcmaps_log_debug (5, "XACML SSL I/O: Got error trigger, but there's no error on the stack\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_ZERO_RETURN :
        {
            /* Shutdown given, possibly not shutdown yet in the underlying transport layer */
            if (printmsg)
                lcmaps_log_debug (5, "XACML SSL I/O: Shutdown given, possibly not shutdown yet in the underlying transport layer\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_WANT_READ:
        case SSL_ERROR_WANT_WRITE:
        {
            if (printmsg)
                lcmaps_log_debug (5, "XACML SSL I/O: operation not completed, retry that last I/O operation\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_WANT_CONNECT:
        case SSL_ERROR_WANT_ACCEPT:
        {
            if (printmsg)
                lcmaps_log_debug (5, "XACML SSL I/O: operation not completed, underlying I/O not ready, retry that last accept/connect operation. Re-check select() or poll() logics\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_WANT_X509_LOOKUP:
        {
            if (printmsg)
                lcmaps_log_debug (5, "XACML SSL I/O: operation not completed, application callback was called again, should retry.\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_SYSCALL:
        {
            /* Some I/O error occurred. The OpenSSL error queue may contain more information on the error. If the error queue is empty (i.e. ERR_get_error() returns 0), ret can be used to find out more about the error: If ret == 0, an EOF was observed that violates the protocol. If ret == -1, the underlying BIO reported an I/O error (for socket I/O on Unix systems, consult errno for details). */
            if (printmsg)
                 lcmaps_log (1, "XACML SSL I/O: Some I/O error occurred. Unrecoverable\n");
            unrecoverable = 1;
            break;
        }
        case SSL_ERROR_SSL:
        {
            /* Check ERR_get_error() for more information */
            if (printmsg)
                lcmaps_log (1, "XACML SSL I/O: A failure in the SSL library occurred, usually a protocol error.\n");
            unrecoverable = 0;
            break;
        }

        default:
        {
            if (printmsg)
                lcmaps_log (1, "XACML SSL I/O: Undefined error code, not in OpenSSL specs\n");
            unrecoverable = 1;
            break;
        }
    }

    if (error_code)
        *error_code = rc;


    return unrecoverable;
}



/*  When the proxy is located on an NFS mount and on the server side the root squash
 *  option has been enabled, the effective uid is mapped to user 'nobody' which should
 *  not be able to read the proxy file. To work around this problem, the effective
 *  uid of the process is changed to that of the calling user and once glexec is done,
 *  the saved uid is used to restore the identity of the process,
 */
int changeEffectiveToRealUid (uid_t * real_uid, uid_t * saved_uid)
{
    struct passwd * this_pw = NULL;

    /* Seg fault safety */
    if (!real_uid || !saved_uid)
    {
        lcmaps_log (1, "%s: Code error: No storage allocated to the real_uid or saved_uid\n", __func__);
        return 1;
    }

    /* For the moment, this is a global setting that will be used to determine a different account to which the process will need to switch before performing a particular action */
    if (certOwner && strlen(certOwner))
    {
        if ((this_pw = getpwnam(certOwner)))
        {
            *real_uid = this_pw -> pw_uid;
            lcmaps_log (5, "%s: (Temporarily) changing the UID of the process to \'%s(%d)\'\n", __func__, certOwner, *real_uid);
        }
        else
        {
            lcmaps_log (1, "%s: Error: can't lookup the account \'%s\'\n", __func__, certOwner);
            return 1;
        }
    }
    else
    {
        *real_uid = getuid();
    }


    /* Save it */
    *saved_uid = geteuid();
    if (seteuid(*real_uid))
    {
        lcmaps_log (1, "%s: Error on changing the process Effective UID with seteuid() to %d\n", __func__, *real_uid);
        return 1;
    }
    return 0;
}

/*  When the proxy is located on an NFS mount and on the server side the root squash
 *  option has been enabled, the effective uid is mapped to user 'nobody' which should
 *  not be able to read the proxy file. To work around this problem, the effective
 *  uid of the process is changed to that of the calling user and once glexec is done,
 *  the saved uid is used to restore the identity of the process,
 */
int restoreEffectiveToRealUid (uid_t real_uid, uid_t saved_uid)
{
    /*  Do not forget to put back the original effective uid on the process. */
    if (seteuid(saved_uid))
    {
        lcmaps_log (1, "%s: Error on returning seteuid() to the saved UID %d\n", __func__, saved_uid);
        return 1;
    }
    return 0;
}



/******************************************************************************
Function:     get_hostname_to_match_client_cert
Description:
              Gets the hostname string to which the peer certificate must match.
              Is this can not be matched automatically, then this feature can be
              used to do so.
Parameters:
              None
Returns:
     NULL : Nothing was set with set_hostname_to_match_client_cert()
    !NULL : String value that must be used in the SSL handshake check to verify 
            the peer certificate.
******************************************************************************/
char * get_hostname_to_match_client_cert(void)
{
    return ssl_override_hostname;
}


/******************************************************************************
Function:     set_hostname_to_match_client_cert
Description:
              Sets the hostname string to which the peer certificate must match.
              Is this can not be matched automatically, then this feature can be
              used to do so.
Parameters:
     override_h_n : The string, which must be a hostname, that overrides the 
                    hostname used to setup the connection and which will be 
                    matched against the peer-certificate information.
 
Returns:
     0 : Good. The string has been duplicated from 'override_h_n'.
     1 : Bad. Memory allocation has failed.
******************************************************************************/
int set_hostname_to_match_client_cert(char * override_h_n)
{
    if ((ssl_override_hostname = strdup (override_h_n)))
        return 0;
    else
        return 1;
}

/******************************************************************************
Function:     free_hostname_to_match_client_cert
Description:
              Sets the hostname string to which the peer certificate must match.
              Is this can not be matched automatically, then this feature can be
              used to do so.
Parameters:
     override_h_n : The string, which must be a hostname, that overrides the 
                    hostname used to setup the connection and which will be 
                    matched against the peer-certificate information.
 
Returns:
     0 : Good. The string has been duplicated from 'override_h_n'.
     1 : Bad. Memory allocation has failed.
******************************************************************************/
void free_hostname_to_match_client_cert(void)
{
    free(ssl_override_hostname);
    ssl_override_hostname = NULL;
}


