/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */



#include <string.h>
#include <stdio.h>

#include <xacml_client.h>
#include <xacml_server.h>

#include <lcmaps/lcmaps_log.h>

/* Network and SSL */
#include "net_common.h"
#include "ssl-common.h"
/* Network and SSL */
#include <stdlib.h>



/* State of this thread */
typedef struct
{
    int             socket;
    SSL            *ssl;
}
ssl_io_state_t;


/* Prototype */
static int ssl_io_close(void  *arg);


static
void *
ssl_io_accept(
    int                                 msocket,
    struct sockaddr                    *addr,
    socklen_t                          *addr_len,
    int                                *sock_out)
{
    char            *logstr = "ssl_io_accept()";
    int              rc = 0;
    ssl_io_state_t  *state = NULL;
    char *           remote_host = NULL;
    SSL_CTX *        my_ctx      = NULL;


    lcmaps_log_debug (5, "Entering %s\n", logstr);

    /* Allocate thread specific state */
    state = malloc (sizeof (ssl_io_state_t));
    if (!state)
    {
        lcmaps_log (1, "%s: memory allocation of ssl_io_state_t failed\n", logstr);
        rc = -1;
        return NULL;
    }
    else
    {
        state->socket = 0;
        state->ssl    = 0;
    }



    /******* SSL Init *********/
    my_ctx = ssl_ctx;
    if (!my_ctx)
    {
        my_ctx = setup_server_ctx(NULL, certCAPath, certFilePub, certFilePriv);
        if (!my_ctx)
        {
            lcmaps_log (1, "%s: Error: unable to create an SSL_CTX context with setup_server_ctx()\n", logstr);
            goto skip_everything_and_exit;
        }
        else
        {
            lcmaps_log_debug (5, "%s: Created CTX here\n", logstr); 
        } 
    }
    else
    {
        lcmaps_log_debug (5, "%s: Got SSL_CTX object\n", logstr); 
    } 
    /**************************/


    /* Setup of Master Socket to listen on the port */
    if (msocket <= 0)
    {
        lcmaps_log (1, "%s: Failure in master listening socket\n", logstr);
        goto skip_everything_and_exit;
    }


    *addr_len = sizeof(struct sockaddr_in);


    /* Accept and create client socket */
    if ((accept_new_client_socket (msocket, &(state -> socket), (struct sockaddr_in *) addr, (socklen_t *) addr_len, &remote_host)) || (state -> socket < 0))
    {
        lcmaps_log (1, "%s: Failed to accept the client socket connection!\n", logstr);
        rc = 14;
        goto finalize_TCPIP;
    }
    else
    {
        /* SSL stuff */
        if ((state -> ssl = SSL_server_accept (state -> socket,
                                               remote_host,
                                               my_ctx
                                      )) == NULL)
        {
            lcmaps_log (1, "%s: Failure in SSL layer setup and connection!\n", logstr);
            rc = 15;
            goto finalize_TCPIP;
        }
        else
        {
            lcmaps_log (5, "%s: SSL layer (server side) connection accepted from host %s\n", logstr, remote_host);
            goto succesful_accept;
        }
        /* SSL stuff */
    }



/*** Good exit ***/
succesful_accept:
    /* Return the connected client socket to the XACML library to listen for traffic */
    if (remote_host) free(remote_host);
    *sock_out = state -> socket;
    lcmaps_log_debug (5, "%s: done\n", logstr);
    return state;


/*** Bad exit ***/
/* finalize_SSL: */
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (SSL_connect_shutdown (state -> ssl))
        {
            lcmaps_log (1, "%s: Error in Server SSL_connect_shutdown!\n", logstr);
        }
        SSL_free(state -> ssl);
    }

finalize_TCPIP:
    if (state)
    {
        close (state -> socket);
    }

skip_everything_and_exit:
    if (state)
    {
        free(state);
        state = NULL;
    }

    if (remote_host) free(remote_host);
    lcmaps_log (3, "%s: ended in a failure\n", logstr);
    return NULL;
}




static 
void *
ssl_io_connect(
    const char                         *endpoint,
    const char                         *host,
    int                                 port)
{
    char            *logstr                = "ssl_io_connect()";
#ifdef OVERRIDE_API_HOST_CHECK
    char            *override_hostname         = NULL;
#endif
    ssl_io_state_t  *state = NULL;

    int              realtime_timeout      = 1; /* seconds */
    int              backoff_milli_seconds = 50;  /* milli seconds */
    int              socket_timeout        = 170; /* milli seconds */

    int              err     = 0;
    int              rc      = 0;
    time_t           start   = 0;
    unsigned int     seed    = 0;


    lcmaps_log_debug (5, "Entering %s\n", logstr);

    /* Initialize OpenSSL Library, as it might not have been initialized before */
    init_OpenSSL();

    /* Random function seeding */
    seed = (unsigned int)time((time_t *)NULL);
    srand(seed);
    /* Random function seeding */

    lcmaps_log_debug (5, "%s: realtime_timeout (%d), backoff_milli_seconds (%d), socket_timeout (%d)\n", 
                        logstr, realtime_timeout, backoff_milli_seconds, socket_timeout);

    /* Allocate thread specific state */
    state = malloc (sizeof (ssl_io_state_t));
    if (!state)
    {
        lcmaps_log (1, "%s: memory allocation of ssl_io_state_t failed\n", logstr);
        rc = -1;
        return NULL;
    }
    else
    {
        state->socket = 0;
        state->ssl    = 0;
    }


    start = time(NULL);
    while ((time(NULL) - start) < realtime_timeout)
    {
        /* Create client socket */
        lcmaps_log_debug (5, "%s: Connecting socket to host %s on port %d\n", logstr, host, port);
        if (create_client_socket (&(state -> socket),
                                   host,
                                   port,
                                   socket_timeout))
        {
            /* Retry connection setup, with a random wait */
            lcmaps_log_debug (2, "%s: Socket creation failed, retrying... in %d milli seconds!\n", logstr, backoff_milli_seconds);

            /* Exponential back off in microseconds */
            usleep (backoff_milli_seconds * 1000);
            backoff_milli_seconds *= 2;
            backoff_milli_seconds += rand_r (&seed) % 25;
            continue; /* Continue until the timeout is reached */
        }
        else
        {
            /* TCP/IP fully connected */
            lcmaps_log_debug (5, "%s: Connected on TCP/IP\n", logstr);
            lcmaps_log_debug (3, "%s: Is connected to host '%s' on port '%d'\n", logstr, host, port);

            /* SSL */
            if ((state -> ssl = SSL_client_connect (state -> socket,
                                                    setup_client_ctx (NULL,
                                                                      certCAPath,
                                                                      certFilePub,
                                                                      certFilePriv)
                                          )) == NULL)
            {
                lcmaps_log (1, "%s: Failure in SSL layer setup and connection!\n", logstr);

                /* Finalize TCP/IP */
                rc = 15;
                goto finalize_TCPIP;
            }
            else
            {
                lcmaps_log_debug (5, "%s: Connected on SSL!\n", logstr);

                /* Post SSL connection check */
#ifdef OVERRIDE_API_HOST_CHECK
                /* Check if there is a hostname override set, if not, perform regular checks */
                if ((override_hostname = get_hostname_to_match_client_cert()))
                {
                    lcmaps_log_debug (5, "%s: Overriding hostname check from \'%s\' to \'%s\'\n", logstr, host, override_hostname);
                    err = post_connection_check(state -> ssl, override_hostname);

                    /* Purge allocated memory, as it has served its purpose */
                    free_hostname_to_match_client_cert();
                }
                else
                {
                    err = post_connection_check(state -> ssl, host);
                }
#else
                /* No override possible, the SSL peer certificate must match the 'host' value */
                err = post_connection_check(state -> ssl, host);
#endif
                if (err != X509_V_OK)
                {
                    lcmaps_log (2, "%s: Error with peer certificate after the full handshake: %s\n", logstr, X509_verify_cert_error_string(err));
                    /* SSL_set_verify_result (state -> ssl, err);  */

                    /* Finalize SSL connection */
                    rc = 16;
                    goto finalize_SSL;
                }
                else
                {
                    lcmaps_log_debug (5, "%s: SSL layer (client side) connection done\n", logstr);

                    /* Finalize SSL connection */
                    rc = 0;
                    goto succesful_connect;
                }
            }
            /* SSL */
        }
    }


    /* Failure */
    lcmaps_log_debug (2, "%s: Failed to connect to host '%s' on port '%d'\n", logstr, host, port);
    goto finalize_TCPIP;


/* Good exit */
succesful_connect:
    lcmaps_log_debug (5, "%s: done\n", logstr);
    return state;


/* Bad exit */
finalize_SSL:
    ssl_io_close (state);
    lcmaps_log_debug (3, "%s: ended in a failure\n", logstr);
    return NULL;

    /* There should be an SSL object to destroy in the first place */
#if 0
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (SSL_connect_shutdown (state -> ssl))
        {
            lcmaps_log_debug (2, "%s: Error: in client SSL_connect_shutdown!\n", logstr);
        }
        SSL_free(state -> ssl);
    }
#endif

finalize_TCPIP:
    if (state)
    {
        close (state -> socket);
        free(state);
        state = NULL;
    }

    lcmaps_log_debug (3, "%s: ended in a failure\n", logstr);
    return NULL;
}
/* ssl_io_connect() */


static
int
ssl_io_send(
    void                               *arg,
    const char                         *data,
    size_t                              size)
{
    char *                              logstr = "ssl_io_send()";
    ssl_io_state_t *                    state = arg;
    int                                 rc = -2;
    int                                 nwritten = 0;

    lcmaps_log_debug (5, "%s: Sending...\n", logstr);

    if (state && state -> ssl)
    {
        for (nwritten = 0; nwritten < size; nwritten += rc)
        {
            rc = SSL_write (state -> ssl, data + nwritten, size - nwritten);
            /* Error conditions are either zero or < zero. */
            if (rc <= 0)
            {
                print_ssl_error_msg (state -> ssl, rc, 1, NULL);
                return -1;
            }
        }
    }


    BIO_flush(state -> ssl -> wbio);

    lcmaps_log_debug (5, "%s: Send %d of %d\n", logstr, rc, size);  
    return 0;
}
/* ssl_io_send() */


static
size_t
ssl_io_recv(
    void                               *arg,
    char                               *data,
    size_t                              size)
{
    char *           logstr = "ssl_io_recv()";
    ssl_io_state_t * state = arg;
    int              rc = 0;


    if (state && state -> ssl)
    {
        if ((rc = SSL_recv (state -> ssl, data, size)) < 0) 
        {
            BIO_flush(state -> ssl -> rbio);
            lcmaps_log (2, "%s: Failure in SSL_recv\n", logstr);
            return -1; 
        }
    }

    lcmaps_log_debug (5, "%s: End of ssl_io_recv()\n", logstr);
    return rc;
}
/* ssl_io_recv() */



static
int
ssl_io_close(
             void  *arg)
{
    char *                          logstr = "ssl_io_close()";
    ssl_io_state_t *                state = arg;

    lcmaps_log_debug (5, "%s: Entering ssl_io_close()\n", logstr);

    /* Only work with it, if not already destroyed or set to NULL */
    if (state)
    {
        if (state -> ssl)
        {    
            /* Shutdown active SSL layer */
            if (SSL_connect_shutdown (state -> ssl))
            {
                lcmaps_log_debug (2, "%s: Error: in SSL_connect_shutdown!\n", logstr);
            }
            state -> ssl -> rbio = NULL;
            state -> ssl -> wbio = NULL;
            lcmaps_log_debug (3, "%s: going to SSL_free()\n", logstr);
            
            SSL_free(state -> ssl);
        }
        close (state -> socket);

        free(state);
        state = NULL;
    }


    lcmaps_log_debug (5, "%s: ended perfectly.\n", logstr);
    return 0;
}
/* ssl_io_close() */


xacml_io_descriptor_t
xacml_io_ssl_descriptor =
{
    "xacml_io_ssl_descriptor",
    ssl_io_accept,
    ssl_io_connect,
    ssl_io_send,
    ssl_io_recv,
    ssl_io_close
};

